/* tub_mmap.c -- allocate page frames for mmap(0,...): evade TASK_UNMAPPED_BASE.
   Copyright 2003 BitWagon Software LLC.  All rights reserved.
   This file may be copied, modified, and distributed under the terms
   of the GNU General Public License (GPL) version 2.
*/
#include <sys/mman.h>
#include <sys/user.h>
#include <asm/unistd.h>
#include <errno.h>
typedef unsigned short ushort;

void *task_unmapped_base;

#define LG2(x) ((1>=(x)) ? 0 : (2>=(x)) ? 1 : (4>=(x)) ? 2 : (8>=(x)) ? 3 : 4)

static unsigned lock_bitmap = 0;  /* multi-thread mutual exclusion */
/* One bit per page.  There are only (1<<20) pages. */
static unsigned long bitmap[1+ ((unsigned long)(void *)~0ul >>
    (LG2(sizeof(unsigned long)) + 3 + PAGE_SHIFT) )];

static
unsigned
czlo(unsigned long w)
{
    int n = 0;
    if (0==              w ) {     return 32; }
    if (0==(0x0000ffff & w)) { w>>=16; n+=16; }
    if (0==(0x000000ff & w)) { w>>= 8; n+= 8; }
    if (0==(0x0000000f & w)) { w>>= 4; n+= 4; }
    if (0==(0x00000003 & w)) { w>>= 2; n+= 2; }
    if (0==(0x00000001 & w)) {         n+= 1; }
    return n;
}

extern int tub_lock(unsigned *);

void
free_pages(void const *const addr, size_t const len)
{
    unsigned const wshift = 3+ LG2(sizeof(bitmap[0]));
    void const *const addr2 = ~PAGE_MASK + len + addr;
    unsigned bot     =  ((unsigned long)addr )>>PAGE_SHIFT;
    unsigned const n = (((unsigned long)addr2)>>PAGE_SHIFT) - bot;
    unsigned const b = ~(~0ul<<wshift) & bot;
    unsigned       w = bot>>wshift;
    int need = b + n;
    unsigned long t;

    tub_lock(&lock_bitmap);
    for (
        t= ~(~0ul<<b) & bitmap[w];  /* first [partial] word */
        0 < (need -= (1<<wshift));
        t= 0ul  /* full words */
    ) {
        bitmap[w++] = t;  /* middle words */
    }
    if (0!=need) { /* partial word at end */
        need &= ~(~0<<wshift);
        t |= (~0ul<<need);
    }
    bitmap[w] &= t;  /* last word */
    tub_unlock(&lock_bitmap);
}

void
claim_pages(void const *const addr, size_t const len, int const need_lock)
{
    unsigned const wshift = 3+ LG2(sizeof(bitmap[0]));
    void const *const addr2 = ~PAGE_MASK + len + addr;
    unsigned bot     =  ((unsigned long)addr )>>PAGE_SHIFT;
    unsigned const n = (((unsigned long)addr2)>>PAGE_SHIFT) - bot;
    unsigned const b = ~(~0ul<<wshift) & bot;
    unsigned       w = bot>>wshift;
    int need = b + n;
    unsigned long t;

    if (need_lock) {
        tub_lock(&lock_bitmap);
    }
    for (
        t= (~0ul<<b) | bitmap[w];  /* first [partial] word */
        0 < (need -= (1<<wshift));
        t= ~0ul  /* full words */
    ) {
        bitmap[w++] = t;  /* middle words */
    }
    if (0!=need) { /* partial word at end */
        need &= ~(~0<<wshift);
        t &= ~(~0ul<<need);
    }
    bitmap[w] |= t;  /* last word */
    if (need_lock) {
        tub_unlock(&lock_bitmap);
    }
}

unsigned
alloc_pages(void const *const addr, size_t const len, void const *const base)
{
    unsigned const wshift = 3+ LG2(sizeof(bitmap[0]));
    void const *const addr2 = ~PAGE_MASK + len + addr;
    unsigned bot     =  ((unsigned long)addr )>>PAGE_SHIFT;
    unsigned const n = (((unsigned long)addr2)>>PAGE_SHIFT) - bot;
    int need = n;
    bot =               ((unsigned long)base )>>PAGE_SHIFT;

    tub_lock(&lock_bitmap);
    while ((n + bot)<=(1+ (~0ul>>PAGE_SHIFT))) {
        unsigned const b = ~(~0ul<<wshift) & bot;
        unsigned       w = bot>>wshift;
        unsigned long t;
        need = b + n;

            /* find contiguous 0 bits */
        for (
            t= (~0ul<<b) & bitmap[w];
            0 < need;
            t= bitmap[++w]
        ) {
            if (0==t) { /* whole word is available */
                need -= 1<<wshift;
            }
            else { /* only part word is available */
                need -= (t= czlo(t));
                break;
            }
        }
        if (need<=0) {
            break;
        }

            /* skip contiguous 1 bits */
        bot = w<<wshift;
        for (
            t= (~(~0ul<<t)) | bitmap[w];  /* first [partial] word */
            ~0ul==t;
            t= bitmap[++w]
        ) {
            bot += 1<<wshift;  /* full words */
        }
        bot += czlo(~t);  /* last [partial] word */
    }
    if (0 < need) { /* failure */
        bot = ~0ul;
    }
    else { /* success */
        void const *const addr3 = (void *)((~PAGE_MASK & (unsigned long)addr) |
            (bot<<PAGE_SHIFT) );
        claim_pages(addr3, len, 0);
    }
    tub_unlock(&lock_bitmap);
    return bot;
}

struct mmap_args {
    void *addr;
    size_t len;
    int prot;
    int flags;
    int fd;
    off_t offset;
};

unsigned
tub_pre_mmap(int edx, int ecx, int ra1, int ra2,
    void *volatile addr, size_t len, int prot, int volatile flags
    /*, int fd, off_t offset*/
)
{
    if (!(MAP_FIXED & flags)) {
        unsigned const frame = alloc_pages(addr, len, task_unmapped_base);
        flags |= MAP_FIXED;
        addr = (void *)((~PAGE_MASK & (unsigned long)addr) |
            (frame<<PAGE_SHIFT) );
        /* failure of alloc_pages() will propagate to failure of mmap() */
    }
    else {
        claim_pages(addr, len, 1);
    }
    return __NR_mmap;
}

unsigned
tub_pre_mmap64( /* register args via PUSHA */
    /* EDI */ int fd,
    /* ESI */ int volatile flags,
    /* EBP */ unsigned page_number,
    /* ESP */ int esp,
    /* EBX */ void *volatile addr,
    /* EDX */ int prot,
    /* ECX */ size_t len
)
{
    if (!(MAP_FIXED & flags)) {
        unsigned const frame = alloc_pages(addr, len, task_unmapped_base);
        flags |= MAP_FIXED;
        addr = (void *)((~PAGE_MASK & (unsigned long)addr) |
            (frame<<PAGE_SHIFT) );
        /* failure of alloc_pages() will propagate to failure of mmap() */
    }
    else {
        claim_pages(addr, len, 1);
    }
    return __NR_mmap2;
}

#define GS 0x65

#define RR  3
#define D32 2
#define D8  1
#define PTR 0
#define ABS 0  /* with EBP==b */

#define EAX 0
#define ECX 1
#define EDX 2
#define EBX 3
#define ESP 4
#define EBP 5
#define ESI 6
#define EDI 7

#define MAB(m,a,b) (((m)<<6)|((a)<<3)|(b))
#define SIB(s,i,b) (((s)<<6)|((i)<<3)|(b))

#define CALL  0xE8
#define JMP32 0xE9
#define STORE 0x89
#define FETCH 0x8B
#define LEA   0x8D
#define MOVI  0xB8
#define POPR  0x58
#define PUSHR 0x50
#define CMPA  0x3D

extern int __pre_mmap(void);
extern int __post_mmap(void);
extern int __pre_mmap64(void);
extern int __post_mmap64(void);

extern int __post_munmap(void);
#include <signal.h>

unsigned __attribute((regparm(1)))
tub_post_mmap(unsigned rv, int edx, int ecx, int ra1, int ra2,
    void *addr, size_t len, int prot, int flags
    /*, int fd, off_t offset*/
)
{
    if (rv==(PAGE_MASK & (unsigned long)addr)) { /* success */
        struct sigaction oldact;
        setup_sigbus_handler(&oldact, tub_post_mmap, &&next);
        if (0!=(PROT_EXEC     & prot)
        &&  0!=(MAP_PRIVATE   & flags)
        &&  0==(MAP_ANONYMOUS & flags) ) {
            unsigned char *const end = len -18 + addr;
            unsigned char *p = addr;
                    /* This is very fragile.  [Making it robust is WORK.] */
            for (; p < end; ++p) if ((EAX+MOVI)==*p) {
                if (                    __NR_mmap==*   (int *)( 1+p)
                &&  ((4<<24)|(SIB(0,ESP,ESP)<<16)|(MAB(D8,EBX,ESP)<<8)|LEA)
                                                 ==   *(int *)( 5+p)
                &&   ((MAB(RR,EBX,EDX)<<8)|STORE)==*(ushort *)(p-2)
                &&                         0x80CD==*(ushort *)( 9+p)
                &&   ((MAB(RR,EDX,EBX)<<8)|STORE)==*(ushort *)(11+p)
                &&                                     CMPA==*(13+p)
                &&          0xfffff000==(0xfffff000 & *(int *)(14+p)) ) {
                    tub_iprotect(p, 18, PROT_WRITE|PROT_READ);
                    p[ 0] = CALL;
                    *(int *)( 1+p) = (unsigned char *)&__pre_mmap  - ( 5+p);

                    p[13] = CALL;
                    *(int *)(14+p) = (unsigned char *)&__post_mmap - (18+p);
                    tub_iprotect(p, 18, prot);
                    p += (18 -1);
                } else
                if (                  __NR_munmap==*   (int *)( 1+p)
                &&                         0x80CD==*(ushort *)( 5+p)
                &&   ((MAB(RR,EDX,EBX)<<8)|STORE)==*(ushort *)( 7+p)
                &&                                     CMPA==*( 9+p)
                &&          0xfffff000==(0xfffff000 & *(int *)(10+p)) ) {
                    p += 9;
        tpmm_munmap:
                    tub_iprotect(p, 5, PROT_WRITE|PROT_READ);
                    p[0] = CALL;
                    *(int *)(1+p) = (unsigned char *)&__post_munmap - (5+p);
                    tub_iprotect(p, 5, prot);
                    p += (5 -1);
                } else
                if (                  __NR_munmap==*   (int *)( 1+p)
                &&                             GS==*          ( 5+p)
                &&     ((MAB(ABS,2,EBP)<<8)|0xFF)==*(ushort *)( 6+p)
                &&                           0x10==*   (int *)( 8+p)
                &&   ((MAB(RR,EDX,EBX)<<8)|STORE)==*(ushort *)(12+p)
                &&                                     CMPA==*(14+p)
                &&          0xfffff000==(0xfffff000 & *(int *)(15+p)) ) {
                    p += 14;
                    goto tpmm_munmap;
                } else
                if (                   __NR_mmap2==*   (int *)( 1+p)
                &&                         0x80CD==*(ushort *)( 5+p)
                &&                                 POPR==(~7&*( 7+p))
                &&                                 POPR==(~7&*( 8+p))
                &&                                 POPR==(~7&*( 9+p))
                &&                                 POPR==(~7&*(10+p))
                &&                                     CMPA==*(11+p)
                &&          0xfffff000==(0xfffff000 & *(int *)(12+p)) ) {
                    tub_iprotect(p, 16, PROT_WRITE|PROT_READ);
                    p[0] = CALL;
                    *(int *)( 1+p) = (unsigned char *)&__pre_mmap64  - ( 5+p);
    
                    *(int *)(12+p) = *(int *)(7+p);  /* all 4 POPR */

                    p[7] = CALL;
                    *(int *)( 8+p) = (unsigned char *)&__post_mmap64 - (12+p);
                    tub_iprotect(p, 16, prot);
                    p += (16 -1);
                } else
                if (                  __NR_mmap2==*   (int *)( 1+p)
                &&                            GS==*          ( 5+p)
                &&    ((MAB(ABS,2,EBP)<<8)|0xFF)==*(ushort *)( 6+p)
                &&                          0x10==*   (int *)( 8+p)
                &&                                POPR==(~7&*(12+p))
                &&                                POPR==(~7&*(13+p))
                &&                                POPR==(~7&*(14+p))
                &&                                POPR==(~7&*(15+p))
                &&                                    CMPA==*(16+p)
                &&         0xfffff000==(0xfffff000 & *(int *)(17+p)) ) {
                    tub_iprotect(p, 21, PROT_WRITE|PROT_READ);
                    p[0] = CALL;
                    *(int *)( 1+p) = (unsigned char *)&__pre_mmap64  - ( 5+p);
    
                    *(int *)(17+p) = *(int *)(12+p);  /* all 4 POPR */

                    p[12] = CALL;
                    *(int *)(13+p) = (unsigned char *)&__post_mmap64 - (17+p);
                    tub_iprotect(p, 21, prot);
                    p += (21 -1);
                }
next: ;
            }
        }
        unset_sigbus_handler(&oldact);
    }
    else { /* failure */
        free_pages(addr, len);
    }
    return rv;
}

unsigned __attribute((regparm(1)))
tub_post_mmap64(unsigned rv,
        /* register args via PUSHA */
    /* EDI */ int fd,
    /* ESI */ int volatile flags,
    /* EBP */ unsigned page_number,
    /* ESP */ int esp,
    /* EBX */ void *volatile addr,
    /* EDX */ int prot,
    /* ECX */ size_t len
)
{
    return tub_post_mmap(rv, 0, 0, 0, 0, addr, len, prot, flags);
}

unsigned __attribute((regparm(1)))
tub_post_munmap(unsigned rv, int edx, int ecx, int ra1, int ra2, void *base, unsigned len)
{
    if (rv <= 0xfffff000) {
        free_pages(base, len);
    }
    return rv;
}

/*
vi:ts=4:et:
*/
